'''
Class definiations are defined based on the CVE JSON V5 Schema: https://github.com/CVEProject/cve-schema/blob/master/schema/v5.0/CVE_JSON_5.0_schema.json

CVE List V5 Github Repo: https://github.com/CVEProject/cvelistV5
'''


'''
CVE Metadata, contains two types:
    1. Published - Required fields: cveId, assignOrgId, state
    2. Rejected - Required fields: cveId, assignOrgId, state
'''
class CveMetadata:
    def __init__(self, **kwargs):
        vars(self).update(kwargs)

    def __str__(self) -> str:
        return str(vars(self))


class CveMetadataPublished(CveMetadata):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)


class CveMetadataRejected(CveMetadata):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)


'''
Container, contains three types:
    1. CnaPublishedContainer - Required fields: providerMetadata, descriptions, affected, references
    2. CnaRejectedContainer - Required fields: providerMetadata, descriptions, affected, references
    3. AdpContainer - Required fields: providerMetadata
'''
class Container:
    def __init__(self, **kwargs):
        vars(self).update(kwargs)

    def __str__(self) -> str:
        return str(vars(self))

    def add_metrics(self, container_type, metrics: "Metrics"):
        # vars(self)[container_type].update({"metrics": vars(metrics)})
        # vars(self)[container_type]["metrics"] = vars(metrics)
        vars(self)[container_type].update({"metrics": metrics})

    def get_metrics(self, container_type):
        return vars(self)[container_type]["metrics"]

    def get_container_type(self):
        if isinstance(self, CnaContainer):
            return "cna"
        elif isinstance(self, AdpContainer):
            return "adp"
        else:
            None


class CnaContainer(Container):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def add_metrics(self, metrics: "Metrics"):
        super().add_metrics("cna", metrics)


class CnaPublishedContainer(CnaContainer):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)


class CnaRejectedContainer(CnaContainer):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)


class AdpContainer(Container):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def add_metrics(self, metrics: "Metrics"):
        super().add_metrics("adp", metrics)


'''
Metrics
'''
class Metrics(dict):
    def __init__(self, from_nvd: bool = False, **kwargs):
        if not from_nvd:
            self.process_metrics_from_json(**kwargs)
        else:
            self.process_metrics_from_nvd(**kwargs)
        dict.__init__(self, vars(self))

    def process_metrics_from_json(self, **kwargs):
        if "cvssV3_1" in kwargs:
            vars(self).update(kwargs["cvssV3_1"])
        elif "cvssV3_0" in kwargs:
            vars(self).update(kwargs["cvssV3_0"])
        elif "cvssV2_0" in kwargs:
            vars(self).update(kwargs["cvssV2_0"])

    def process_metrics_from_nvd(self, **kwargs):
        def process_metrics(metrics):
            del metrics[0].source
            del metrics[0].type
            vars(self).update(vars(metrics[0].cvssData))
            del metrics[0].cvssData
            vars(self).update(vars(metrics[0]))

        if "cvssMetricV31" in kwargs["metrics"]:
            process_metrics(kwargs["metrics"].cvssMetricV31)
        elif "cvssMetricV30" in kwargs["metrics"]:
            process_metrics(kwargs["metrics"].cvssMetricV30)
        elif "cvssMetricV2" in kwargs["metrics"]:
            process_metrics(kwargs["metrics"].cvssMetricV2)

    def __str__(self):
        return str(vars(self))



__all__ = ["CveMetadataPublished", "CveMetadataRejected", "CnaPublishedContainer", "CnaRejectedContainer", "AdpContainer", "Metrics"]



import pathlib


DEFAULT_PROJECT_DIR = pathlib.Path.home() / ".config/cvedb"
DEFAULT_PROJECT_LOCAL_REPO = DEFAULT_PROJECT_DIR / "cvelistV5"

# def home_dir():
#     return pathlib.Path.home()


def create_path(path: str, parents=True, exist_ok=True):
    try:
        p = pathlib.Path(path)
        p.mkdir(mode=0o744, parents=parents, exist_ok=exist_ok)
    except Exception:
        print(f"Exception when creating director {path}")


def path_exists(path: str):
    p = pathlib.Path(path)
    return p.exists()


def open_path(path: str) -> pathlib.Path:
    return pathlib.Path(path)





import re
import json
from tqdm import tqdm

from .cvetools.CVEComponents import *
from .cvetools.CVEHandler import *
from .cvetools.CVEListHandler import CvelistHandler

from .cvedb import *
from .utils import pickleutils
from .utils import pathutils
from .utils import argsutils

from .version import __version__

"""
CVEdb contains Table
Table contains CVEs
table_name is the year of individual CVE

example format can be described as:
{
    '2023': {
        'table_name': "2023",
        'data_count': 2
        'CVE-2023-0001': {},
        'CVE-2023-0002': {}
    },
    '2022': {
        'table_name': "2022",
        'data_count': 2
        'CVE-2022-0001': {},
        'CVE-2022-0002': {}
    }
}
"""


DEFAULT_PATTERN = "**/CVE-*.json"


class CVEdb:
    OUTPUT_PICKLE_FILE = pathutils.DEFAULT_PROJECT_DIR / "cvedb.pickle"

    def __init__(self):
        self.table_count = 0
        self.total_data_count = 0
        self.records: dict[int, Table] = {} # key-value pair, where key is table name, value is table

    def update_stat(self):
        """
        Updates the statistics of the CVEdb object.

        This function calculates and updates the number of tables (or records) and the total data count across all tables.
        The `table_count` is the number of keys in the `records` dictionary.
        The `total_data_count` is calculated by iterating over all tables in `records` and summing up their `data_count` values.

        :return: A tuple containing the table count and the total data count.
        """
        self.table_count = len(self.records.keys())
        count = 0
        for _, v in self.records.items():
            count += v.data_count
        self.total_data_count = count
        return self.table_count, self.total_data_count

    def upsert(self, data: CVE):
        year = data.get_cve_year()
        if year not in self.records:
            self.records[year] = Table(year, 0, {})
        table = self.records[year]
        table.upsert(data)

    def get_cve_by_id(self, cve_id) -> CVE:
        year = int(cve_id.split("-")[1])
        table = self.records[year]
        return table.get_by_id(cve_id)

    def get_cves_by_year(self, year, pattern = None):
        """
        Retrieves all CVEs for a given year that match a certain pattern and returns them in a new Table instance.

        :param year: The year to select the table of CVEs.
        :param pattern: The pattern to filter the CVEs. This is optional.
        :return: A new Table instance containing the CVEs for the given year that match the pattern.
        """
        pattern = argsutils.process_pattern(pattern) if pattern else r"()" # convert cli pattern to regex
        # print(f"Pattern: {pattern}")
        table = self.records[int(year)]
        out = {"table_name": table.table_name, "data_count": 0, "data": {}}
        for k, v in table.data.items():  # k: str, cveid; v: CVE instance
            cve_json = jsonlialize_cve(v)
            if re.match(pattern, str(cve_json)):
                out["data"].update({k: cve_json})
                out["data_count"] = out["data_count"] + 1

        out_table = Table(out["table_name"], out["data_count"], out["data"])  # create a new Table instance
        return out_table

    def __str__(self) -> str:
        self.update_stat()
        return f"Table Count: {self.table_count}\nTotal Data Count: {self.total_data_count}"


class Table:
    def __init__(self, table_name, data_count: int, data: dict[str, CVE]):
        self.table_name = table_name
        self.data_count = data_count
        self.data: dict[str, CVE] = data

    def upsert(self, data: CVE):
        if not data.get_cve_id() in self.data:
            self.data_count += 1
        self.data.update({data.get_cve_id(): data})

    def get_by_id(self, cve_id) -> CVE:
        if not cve_id in self.data:
            raise KeyError("CVE not found")
        return self.data[cve_id]

    def get_data(self):
        return self.data

    def __str__(self):
        return f"Table: {self.table_name}\nData Count: {self.data_count}"


def jsonlialize_cve(data) -> dict:
    out = {}
    for k, v in vars(data).items():
        try:
            json.dumps(v)  # check if the value is json serializable
            out.update({k: v})
        except TypeError:
            out.update({k: jsonlialize_cve(v)})
    return out


def dump_db(cvedb: CVEdb, out_path: str = CVEdb.OUTPUT_PICKLE_FILE):
    """
    Serialize and store the `cvedb` object into a file.

    :param cvedb: The CVEdb object to be stored.
    :param out_path: The path where the serialized object will be stored. Defaults to CVEdb.OUTPUT_PICKLE_FILE.
    """
    print(f"Store cvedb to {out_path}")
    data = pickleutils.compress(pickleutils.serialize(cvedb))
    pickleutils.pickle_dump(out_path, data)


def init_db(db_path = CVEdb.OUTPUT_PICKLE_FILE):
    """
    Initialize a CVE (Common Vulnerabilities and Exposures) database.

    This function tries to load a CVEdb object from a local pickle file. If it cannot find the file or if there is an error during loading, it creates a new CVEdb instance.

    :param db_path: The path where the serialized object is stored. Defaults to CVEdb.OUTPUT_PICKLE_FILE.
    :return: The deserialized CVEdb object or a new CVEdb instance if the file does not exist or there is an error during loading.
    :raises Exception: If there is an error during loading, decompression, or deserialization.
    """
    try:
        print(f"Loading cve database from {db_path}")
        cvedb = pickleutils.pickle_load(db_path)
        cvedb = pickleutils.deserialize(pickleutils.decompress(cvedb))
        return cvedb
    except:
        print(f"No local database found in path {db_path}, creating new CVEdb")
        return CVEdb()


def process_file(file, cve_handler: CVEHandler, create_metrics: bool) -> CVE:
    cve = cve_handler.create_cve_from_json(file)
    if cve.contains_metrics():
        cve.create_metrics(True)  # create Metrics if CVE JSON file contains metrics entry
    else:
        create_metrics and cve.create_metrics(False)
    return cve


def handle_updated_cve(cvedb: CVEdb, local_repo_path: str, files: list, args = None):
    cve_handler = CVEHandler(local_repo_path)
    for f in tqdm(files):
        path = pathutils.DEFAULT_PROJECT_LOCAL_REPO / f
        cve = process_file(path, cve_handler, args.create_metrics)
        cvedb.upsert(cve)


def handle_cve_json(cvedb: CVEdb, local_repo_path: str, pattern: str = DEFAULT_PATTERN, args = None):
    cve_handler = CVEHandler(local_repo_path)
    for f in tqdm(cve_handler.get_cvelist_path().glob(pattern)):
    # for f in cve_handler.get_cvelist_path().glob("**/CVE-2013-3703.json"): # testing purpose, one JSON contains metrics
        cve = process_file(f, cve_handler, args.create_metrics)
        cvedb.upsert(cve)


def clone_or_update(args):
    if args.clone and args.update:
        raise Exception("Invalid arguments combination")
    repo = CvelistHandler()
    cvedb = init_db()
    if args.clone:
        handle_cve_json(cvedb, repo.get_local_repo_path(), args=args)
    elif args.update:
        updated = repo.find_updated_files()
        repo.pull_from_remote()
        handle_updated_cve(cvedb, repo.get_local_repo_path(), files=updated, args=args)
    dump_db(cvedb)


def search(cvedb: CVEdb, year: int, id: str, pattern: str) -> dict | CVE:
    if year:
        return cvedb.get_cves_by_year(year, pattern)
    elif id:
        return cvedb.get_cve_by_id(id)


def main():
    args = argsutils.init_argparse().parse_args()
    # print(vars(args))
    if args.version:
        print(f"CVEdb - {__version__}")
    elif args.clone or args.update:
        clone_or_update(args)
    elif args.search:
        cvedb = init_db()
        data = search(cvedb, args.year, args.id, args.pattern)
        # print(json.dumps(jsonlialize_cve(data), indent=2))
        # print(type(data))
        return data


if __name__ == "__main__":
    main()


# __all__ = ["CVEdb"]




import json

from ..utils import pathutils
from .CVEComponents import *
from ..nvdapi import CVEQuery


class CVE:
    def __init__(self, metadata, containers, data_type = "CVE Record", data_version = "5.0"):
        self.data_type = data_type
        self.data_version = data_version
        self.metadata = metadata
        self.containers = containers

    def __str__(self) -> str:
        return str(vars(self))

    def contains_metrics(self) -> bool:
        def check_metrics(container_type):
            return "metrics" in vars(self.containers)[container_type]

        return check_metrics(self.containers.get_container_type())

    def create_metrics(self, metrics_in_json: bool) -> "Metrics":
        def create_metrics_helper(container_type):
            if not metrics_in_json:
                nvd_info = CVEQuery().get_cve_by_id(self.metadata.cveId)
                return Metrics(True, **vars(nvd_info))
            else:
                return Metrics(**vars(self.containers)[container_type]["metrics"][0])

        self.containers.add_metrics(create_metrics_helper(self.containers.get_container_type()))

    def get_cve_id(self) -> str:
        return self.metadata.cveId

    def get_cve_year(self) -> int:
        info = self.metadata.cveId.split("-")
        return int(info[1])

    def get_metrics(self) -> Metrics:
        """
        Retrieves the metrics associated with the container of a specific type.

        :return: The metrics of the container if they exist, otherwise None.
        """
        containers = vars(self.containers)[self.containers.get_container_type()]
        if "metrics" in containers:
            return containers["metrics"]
        return None  # when no metrics entry found

    def get_cvss_score(self) -> float | None:
        """
        Retrieves the CVSS base score from the metrics of the container.

        :return: The CVSS base score if metrics exist, otherwise None.
        """
        metrics = self.get_metrics()
        return float(metrics["baseScore"]) if metrics else None


class CVEHandler:
    def __init__(self, cvelist_path):
        self.cvelist_path = pathutils.open_path(cvelist_path)
        self.state = None
        self.cveMetadata = None
        self.containers = None

    def get_cvelist_path(self):
        return self.cvelist_path

    def create_cve_from_json(self, json_path: str):
        # print(json_path)
        with open(json_path, "r") as file:
            data = json.load(file)
            try:
                return self.create_cve(data)
            except:
                raise Exception(f"Exception when creating CVE instance on file: {json_path}")

    def parse_cve_metadata(self, **metadata):
        self.state = metadata["state"]
        if self.state == "PUBLISHED":
            self.cveMetadata = CveMetadataPublished(**metadata)
        elif self.state == "REJECTED":
            self.cveMetadata = CveMetadataRejected(**metadata)
        else:
            raise TypeError("Invalid CVE Metadata State")

    def parse_containers(self, **containers):
        if "cna" in containers:
            if self.state == "PUBLISHED":
                self.containers = CnaPublishedContainer(**containers)
            elif self.state == "REJECTED":
                self.containers = CnaRejectedContainer(**containers)
            else:
                raise TypeError("Invalid CVE Metadata State")
        elif "adp" in containers:
            self.containers = AdpContainer(**containers)
        else:
            raise TypeError("Invalid Containers Type")

    def create_cve(self, data) -> CVE:
        try:
            self.parse_cve_metadata(**data["cveMetadata"])
            self.parse_containers(**data["containers"])
        except Exception as e:
            raise e

        cve = CVE(self.cveMetadata, self.containers)
        return cve


__all__ = ["CVEHandler", "CVE"]




import git

from ..utils import pathutils

# CVE List V5 Github Repo: https://github.com/CVEProject/cvelistV5
CVE_LIST_V5_REPO = "git@github.com:CVEProject/cvelistV5.git"

"""
reference: https://stackoverflow.com/questions/51045540/python-progress-bar-for-git-clone
add Clone Progress bar
"""
class CloneProgress(git.RemoteProgress):
    def __init__(self) -> None:
        super().__init__()

    def update(self, *args):
        print(self._cur_line, end="\r")


class CvelistHandler:
    """
    Class to handle cvelistV5 repo
    """
    def __init__(self):
        pathutils.create_path(pathutils.DEFAULT_PROJECT_DIR)
        self.local_repo_path = pathutils.DEFAULT_PROJECT_DIR / "cvelistV5"

        if not pathutils.path_exists(self.local_repo_path):
            print("Cloning Repo...")
            self.clone_to_local()
        self.repo = git.Repo(self.local_repo_path)

    def clone_to_local(self):
        git.Repo.clone_from(CVE_LIST_V5_REPO, self.local_repo_path, progress=CloneProgress())

    def find_updated_files(self):
        origin = self.repo.remotes.origin
        origin.fetch()

        remote_hash = origin.refs.main.commit.hexsha

        updated_file = []
        for file in self.repo.index.diff(remote_hash):
            if "delta" in file.a_path:  # ignore delta.json and deltaLog.json
                continue
            updated_file.append(file.a_path)
        return updated_file

    def pull_from_remote(self):
        origin = self.repo.remotes.origin
        origin.pull()

    def get_local_repo_path(self):
        return str(self.local_repo_path)


__all__ = ["CvelistHandler"]



from setuptools import setup, find_packages
from cvedb.version import __version__

long_desc = ""
with open("./README.md", "r") as file:
    for line in file:
        long_desc += line

setup(
    name="py-cvedb",
    version=__version__,
    packages=find_packages(),
    author="Box Hezi",
    author_email="hezipypi.yixdpu@bumpmail.io",
    description="A simple CVE database written in Python",
    long_description=long_desc,
    long_description_content_type="text/markdown",
    url="https://github.com/BoxHezi/cvedb",
    install_requires=[
        "nvdlib",
        "tqdm",
        "pgzip",
        "GitPython"
    ],
    entry_points={
        'console_scripts': [
            'cvedb = cvedb.cvedb:main'
        ]
    }
)